use crate::model::SectionRef;
use crate::section_interface::SectionInterface;
use actix::{
    Actor, ActorContext, Addr, AsyncContext, Handler, Message, MessageResult, SpawnHandle,
};
use futures_util::TryFutureExt;
use std::{
    future::Future,
    mem::swap,
    sync::{
        atomic::{AtomicI32, Ordering},
        Arc,
    },
    time::Duration,
};
use thiserror::Error;
use tokio::{
    sync::{broadcast, watch},
    time::Instant,
};
use tracing::{debug, trace, warn};

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
#[derive(serde::Deserialize, serde::Serialize)]
pub struct SectionRunHandle(i32);

impl SectionRunHandle {
    pub fn into_inner(self) -> i32 {
        self.0
    }
}

#[derive(Clone, Debug)]
pub enum SectionEvent {
    RunStart(SectionRunHandle, SectionRef),
    RunFinish(SectionRunHandle, SectionRef),
    RunPause(SectionRunHandle, SectionRef),
    RunUnpause(SectionRunHandle, SectionRef),
    RunCancel(SectionRunHandle, SectionRef),
    RunnerPause,
    RunnerUnpause,
}

pub type SectionEventRecv = broadcast::Receiver<SectionEvent>;
type SectionEventSend = broadcast::Sender<SectionEvent>;

const EVENT_CAPACITY: usize = 8;

#[derive(Clone, Debug, PartialEq)]
pub enum SecRunState {
    Waiting,
    Running {
        start_time: Instant,
    },
    Finished,
    Cancelled,
    Paused {
        start_time: Instant,
        pause_time: Instant,
    },
}

#[derive(Clone, Debug)]
pub struct SecRun {
    pub(crate) handle: SectionRunHandle,
    pub(crate) section: SectionRef,
    pub(crate) duration: Duration,
    pub(crate) total_duration: Duration,
    pub(crate) state: SecRunState,
}

impl SecRun {
    fn new(handle: SectionRunHandle, section: SectionRef, duration: Duration) -> Self {
        Self {
            handle,
            section,
            duration,
            total_duration: duration,
            state: SecRunState::Waiting,
        }
    }

    pub fn is_running(&self) -> bool {
        matches!(self.state, SecRunState::Running{..})
    }

    #[allow(dead_code)]
    pub fn is_paused(&self) -> bool {
        matches!(self.state, SecRunState::Paused{..})
    }
}

pub type SecRunQueue = im::Vector<Arc<SecRun>>;

#[derive(Clone, Debug)]
pub struct SecRunnerState {
    pub(crate) run_queue: SecRunQueue,
    pub(crate) paused: bool,
}

impl Default for SecRunnerState {
    fn default() -> Self {
        Self {
            run_queue: SecRunQueue::default(),
            paused: false,
        }
    }
}

pub type SecRunnerStateRecv = watch::Receiver<SecRunnerState>;

struct SectionRunnerInner {
    interface: Arc<dyn SectionInterface + Sync>,
    event_send: Option<SectionEventSend>,
    state_send: watch::Sender<SecRunnerState>,
    delay_future: Option<SpawnHandle>,
    did_change: bool,
}

impl SectionRunnerInner {
    fn send_event(&mut self, event: SectionEvent) {
        if let Some(event_send) = &mut self.event_send {
            match event_send.send(event) {
                Ok(_) => {}
                Err(_closed) => {
                    self.event_send = None;
                }
            }
        }
    }

    fn subscribe_event(&mut self) -> SectionEventRecv {
        match &mut self.event_send {
            Some(event_send) => event_send.subscribe(),
            None => {
                let (event_send, event_recv) = broadcast::channel(EVENT_CAPACITY);
                self.event_send = Some(event_send);
                event_recv
            }
        }
    }

    fn start_run(&mut self, run: &mut Arc<SecRun>) {
        use SecRunState::*;
        let run = Arc::make_mut(run);
        debug!(section_id = run.section.id, "starting running section");
        self.interface
            .set_section_state(run.section.interface_id, true);
        run.state = Running {
            start_time: Instant::now(),
        };
        self.send_event(SectionEvent::RunStart(
            run.handle.clone(),
            run.section.clone(),
        ));
        self.did_change = true;
    }

    fn finish_run(&mut self, run: &mut Arc<SecRun>) {
        let run = Arc::make_mut(run);
        if run.is_running() {
            debug!(section_id = run.section.id, "finished running section");
            self.interface
                .set_section_state(run.section.interface_id, false);
            run.state = SecRunState::Finished;
            self.send_event(SectionEvent::RunFinish(
                run.handle.clone(),
                run.section.clone(),
            ));
            self.did_change = true;
        } else {
            warn!(
                section_id = run.section.id,
                state = debug(&run.state),
                "cannot finish run which is not running"
            );
        }
    }

    fn cancel_run(&mut self, run: &mut Arc<SecRun>) {
        let run = Arc::make_mut(run);
        if run.is_running() {
            debug!(section_id = run.section.id, "cancelling running section");
            self.interface
                .set_section_state(run.section.interface_id, false);
        }
        run.state = SecRunState::Cancelled;
        self.send_event(SectionEvent::RunCancel(
            run.handle.clone(),
            run.section.clone(),
        ));
        self.did_change = true;
    }

    fn pause_run(&mut self, run: &mut Arc<SecRun>) {
        use SecRunState::*;
        let run = Arc::make_mut(run);
        let new_state = match run.state {
            Running { start_time } => {
                debug!(section_id = run.section.id, "pausing running section");
                self.interface
                    .set_section_state(run.section.interface_id, false);
                Paused {
                    start_time,
                    pause_time: Instant::now(),
                }
            }
            Waiting => {
                debug!(section_id = run.section.id, "pausing waiting section");
                Paused {
                    start_time: Instant::now(),
                    pause_time: Instant::now(),
                }
            }
            Finished | Cancelled | Paused { .. } => {
                return;
            }
        };
        run.state = new_state;
        self.send_event(SectionEvent::RunPause(
            run.handle.clone(),
            run.section.clone(),
        ));
        self.did_change = true;
    }

    fn unpause_run(&mut self, run: &mut Arc<SecRun>) {
        use SecRunState::*;
        let run = Arc::make_mut(run);
        match run.state {
            Paused {
                start_time,
                pause_time,
            } => {
                debug!(section_id = run.section.id, "unpausing section");
                self.interface
                    .set_section_state(run.section.interface_id, true);
                run.state = Running {
                    start_time: Instant::now(),
                };
                let ran_for = pause_time - start_time;
                run.duration -= ran_for;
                self.send_event(SectionEvent::RunUnpause(
                    run.handle.clone(),
                    run.section.clone(),
                ));
            }
            Waiting | Finished | Cancelled | Running { .. } => {
                warn!(
                    section_id = run.section.id,
                    state = debug(&run.state),
                    "can only unpause paused section"
                );
            }
        }
        self.did_change = true;
    }

    fn process_after_delay(
        &mut self,
        after: Duration,
        ctx: &mut <SectionRunnerActor as Actor>::Context,
    ) {
        let delay_future = ctx.notify_later(Process, after);
        if let Some(old_future) = self.delay_future.replace(delay_future) {
            ctx.cancel_future(old_future);
        }
    }

    fn cancel_process(&mut self, ctx: &mut <SectionRunnerActor as Actor>::Context) {
        if let Some(old_future) = self.delay_future.take() {
            ctx.cancel_future(old_future);
        }
    }
}

struct SectionRunnerActor {
    state: SecRunnerState,
    inner: SectionRunnerInner,
}

impl Actor for SectionRunnerActor {
    type Context = actix::Context<Self>;

    fn started(&mut self, _ctx: &mut Self::Context) {
        trace!("section_runner starting");
    }

    fn stopped(&mut self, _ctx: &mut Self::Context) {
        trace!("section_runner stopped");
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct Quit;

impl Handler<Quit> for SectionRunnerActor {
    type Result = ();

    fn handle(&mut self, _msg: Quit, ctx: &mut Self::Context) -> Self::Result {
        ctx.stop();
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct QueueRun(SectionRunHandle, SectionRef, Duration);

impl Handler<QueueRun> for SectionRunnerActor {
    type Result = ();

    fn handle(&mut self, msg: QueueRun, ctx: &mut Self::Context) -> Self::Result {
        let QueueRun(handle, section, duration) = msg;

        let run: Arc<SecRun> = SecRun::new(handle, section, duration).into();
        self.state.run_queue.push_back(run);
        self.inner.did_change = true;

        ctx.notify(Process);
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct CancelRun(SectionRunHandle);

impl Handler<CancelRun> for SectionRunnerActor {
    type Result = ();

    fn handle(&mut self, msg: CancelRun, ctx: &mut Self::Context) -> Self::Result {
        let CancelRun(handle) = msg;
        for run in self.state.run_queue.iter_mut() {
            if run.handle != handle {
                continue;
            }
            trace!(handle = handle.0, "cancelling run by handle");
            self.inner.cancel_run(run);
        }
        ctx.notify(Process);
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct CancelAll;

impl Handler<CancelAll> for SectionRunnerActor {
    type Result = ();

    fn handle(&mut self, _msg: CancelAll, ctx: &mut Self::Context) -> Self::Result {
        let mut old_runs = SecRunQueue::new();
        swap(&mut old_runs, &mut self.state.run_queue);
        trace!(count = old_runs.len(), "cancelling all runs");
        for mut run in old_runs {
            self.inner.cancel_run(&mut run);
        }
        ctx.notify(Process);
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct SetPaused(bool);

impl Handler<SetPaused> for SectionRunnerActor {
    type Result = ();

    fn handle(&mut self, msg: SetPaused, ctx: &mut Self::Context) -> Self::Result {
        let SetPaused(pause) = msg;
        if pause != self.state.paused {
            if pause {
                self.state.paused = true;
                self.inner.send_event(SectionEvent::RunnerPause);
            } else {
                self.state.paused = false;
                self.inner.send_event(SectionEvent::RunnerUnpause);
            }
            self.inner.did_change = true;
            ctx.notify(Process);
        }
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "SectionEventRecv")]
struct Subscribe;

impl Handler<Subscribe> for SectionRunnerActor {
    type Result = MessageResult<Subscribe>;

    fn handle(&mut self, _msg: Subscribe, _ctx: &mut Self::Context) -> Self::Result {
        let event_recv = self.inner.subscribe_event();
        MessageResult(event_recv)
    }
}

#[derive(Message, Debug, Clone)]
#[rtype(result = "()")]
struct Process;

impl Handler<Process> for SectionRunnerActor {
    type Result = ();

    fn handle(&mut self, _msg: Process, ctx: &mut Self::Context) -> Self::Result {
        self.process(ctx)
    }
}

impl SectionRunnerActor {
    fn new(
        interface: Arc<dyn SectionInterface + Sync>,
        state_send: watch::Sender<SecRunnerState>,
    ) -> Self {
        Self {
            state: SecRunnerState::default(),
            inner: SectionRunnerInner {
                interface,
                event_send: None,
                state_send,
                delay_future: None,
                did_change: false,
            },
        }
    }

    fn process_queue(&mut self, ctx: &mut actix::Context<Self>) {
        use SecRunState::*;
        let state = &mut self.state;
        while let Some(current_run) = state.run_queue.front_mut() {
            let run_finished = match (&current_run.state, state.paused) {
                (Waiting, false) => {
                    self.inner.start_run(current_run);
                    self.inner.process_after_delay(current_run.duration, ctx);
                    false
                }
                (Running { start_time }, false) => {
                    let time_to_finish = start_time.elapsed() >= current_run.duration;
                    if time_to_finish {
                        self.inner.finish_run(current_run);
                        self.inner.cancel_process(ctx);
                    }
                    time_to_finish
                }
                (Paused { .. }, false) => {
                    self.inner.unpause_run(current_run);
                    self.inner.process_after_delay(current_run.duration, ctx);
                    false
                }
                (Waiting, true) => {
                    self.inner.pause_run(current_run);
                    false
                }
                (Running { .. }, true) => {
                    self.inner.pause_run(current_run);
                    self.inner.cancel_process(ctx);
                    false
                }
                (Paused { .. }, true) => false,
                (Cancelled, _) | (Finished, _) => true,
            };

            if run_finished {
                state.run_queue.pop_front();
            } else {
                break;
            }
        }
    }

    fn process(&mut self, ctx: &mut actix::Context<Self>) {
        self.process_queue(ctx);

        if self.inner.did_change {
            let _ = self.inner.state_send.broadcast(self.state.clone());
            self.inner.did_change = false;
        }
    }
}

#[derive(Debug, Clone, Error)]
#[error("error communicating with SectionRunner: {0}")]
pub struct Error(#[from] actix::MailboxError);

pub type Result<T, E = Error> = std::result::Result<T, E>;

#[derive(Clone)]
pub struct SectionRunner {
    state_recv: SecRunnerStateRecv,
    addr: Addr<SectionRunnerActor>,
    next_run_id: Arc<AtomicI32>,
}

#[allow(dead_code)]
impl SectionRunner {
    pub fn new(interface: Arc<dyn SectionInterface + Sync>) -> Self {
        let (state_send, state_recv) = watch::channel(SecRunnerState::default());
        let addr = SectionRunnerActor::new(interface, state_send).start();
        Self {
            state_recv,
            addr,
            next_run_id: Arc::new(AtomicI32::new(1)),
        }
    }

    pub fn quit(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(Quit).map_err(From::from)
    }

    fn queue_run_inner(
        &mut self,
        section: SectionRef,
        duration: Duration,
    ) -> (QueueRun, SectionRunHandle) {
        let run_id = self.next_run_id.fetch_add(1, Ordering::SeqCst);
        let handle = SectionRunHandle(run_id);
        (QueueRun(handle.clone(), section, duration), handle)
    }

    pub fn do_queue_run(
        &mut self,
        section: SectionRef,
        duration: Duration,
    ) -> SectionRunHandle {
        let (queue_run, handle) = self.queue_run_inner(section, duration);
        self.addr.do_send(queue_run);
        handle
    }

    pub fn queue_run(
        &mut self,
        section: SectionRef,
        duration: Duration,
    ) -> impl Future<Output = Result<SectionRunHandle>> {
        let (queue_run, handle) = self.queue_run_inner(section, duration);
        self.addr
            .send(queue_run)
            .map_err(From::from)
            .map_ok(move |_| handle)
    }

    pub fn do_cancel_run(&mut self, handle: SectionRunHandle) {
        self.addr.do_send(CancelRun(handle))
    }

    pub fn cancel_run(&mut self, handle: SectionRunHandle) -> impl Future<Output = Result<()>> {
        self.addr.send(CancelRun(handle)).map_err(From::from)
    }

    pub fn cancel_all(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(CancelAll).map_err(From::from)
    }

    pub fn pause(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(SetPaused(true)).map_err(From::from)
    }

    pub fn unpause(&mut self) -> impl Future<Output = Result<()>> {
        self.addr.send(SetPaused(false)).map_err(From::from)
    }

    pub fn subscribe(&mut self) -> impl Future<Output = Result<SectionEventRecv>> {
        self.addr.send(Subscribe).map_err(From::from)
    }

    pub fn state_receiver(&self) -> SecRunnerStateRecv {
        self.state_recv.clone()
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::section_interface::MockSectionInterface;
    use crate::{
        model::{Section, Sections},
        trace_listeners::{EventListener, Filters},
    };
    use assert_matches::assert_matches;
    use im::ordmap;
    use tracing_subscriber::prelude::*;

    #[actix_rt::test]
    async fn test_quit() {
        let quit_msg = EventListener::new(
            Filters::new()
                .target("sprinklers_rs::section_runner")
                .message("section_runner stopped"),
        );
        let subscriber = tracing_subscriber::registry().with(quit_msg.clone());
        let _sub = tracing::subscriber::set_default(subscriber);

        let interface = MockSectionInterface::new(6);
        let mut runner = SectionRunner::new(Arc::new(interface));
        tokio::task::yield_now().await;
        runner.quit().await.unwrap();

        assert_eq!(quit_msg.get_count(), 1);
    }

    fn make_sections_and_interface() -> (Sections, Arc<MockSectionInterface>) {
        let interface = Arc::new(MockSectionInterface::new(2));
        let sections: Sections = ordmap![
            1 => Section {
                id: 1,
                name: "Section 1".into(),
                interface_id: 0,
            }.into(),
            2 => Section {
                id: 2,
                name: "Section 2".into(),
                interface_id: 1,
            }.into()
        ];
        (sections, interface)
    }

    fn assert_section_states(interface: &MockSectionInterface, states: &[bool]) {
        for (id, state) in states.iter().enumerate() {
            assert_eq!(
                interface.get_section_state(id as u32),
                *state,
                "section interface id {} did not match",
                id
            );
        }
    }

    async fn advance(dur: Duration) {
        // HACK: advance should really be enough, but we need another yield_now
        tokio::time::pause();
        tokio::time::advance(dur).await;
        tokio::task::yield_now().await;
        tokio::time::resume();
    }

    #[actix_rt::test]
    async fn test_queue() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        assert_section_states(&interface, &[false, false]);

        // Queue single section, make sure it runs
        runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[false, false]);

        // Queue two sections, make sure they run one at a time
        runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[true, false]);

        advance(Duration::from_secs(10)).await;

        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_cancel_run() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let run1 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run2 = runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;

        assert_section_states(&interface, &[false, true]);

        runner.cancel_run(run1).await.unwrap();
        tokio::task::yield_now().await;

        assert_section_states(&interface, &[true, false]);

        runner.cancel_run(run3).await.unwrap();
        advance(Duration::from_secs(11)).await;

        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_cancel_all() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.cancel_all().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_pause() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let _run1 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let _run3 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        advance(Duration::from_secs(10)).await;
        assert_section_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(8)).await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(2)).await;
        assert_section_states(&interface, &[true, false]);

        runner.pause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        // cancel paused run
        runner.cancel_run(run2).await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, false]);

        runner.unpause().await.unwrap();
        tokio::task::yield_now().await;
        assert_section_states(&interface, &[false, true]);

        advance(Duration::from_secs(11)).await;
        assert_section_states(&interface, &[false, false]);

        runner.quit().await.unwrap();
    }

    #[actix_rt::test]
    async fn test_event() {
        let (sections, interface) = make_sections_and_interface();
        let mut runner = SectionRunner::new(interface.clone());

        let mut event_recv = runner.subscribe().await.unwrap();

        let run1 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run2 = runner
            .queue_run(sections[&1].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        let run3 = runner
            .queue_run(sections[&2].clone(), Duration::from_secs(10))
            .await
            .unwrap();

        assert_matches!(
            event_recv.recv().await,
            Ok(SectionEvent::RunStart(handle, _))
            if handle == run1
        );

        runner.pause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerPause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunPause(handle, _)) if handle == run1);

        runner.unpause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerUnpause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunUnpause(handle, _)) if handle == run1);

        advance(Duration::from_secs(11)).await;
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunFinish(handle, _)) if handle == run1);
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunStart(handle, _)) if handle == run2);

        runner.pause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerPause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunPause(handle, _)) if handle == run2);

        // cancel paused run
        runner.cancel_run(run2.clone()).await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunCancel(handle, _)) if handle == run2);
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunPause(handle, _)) if handle == run3);

        runner.unpause().await.unwrap();
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunnerUnpause));
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunUnpause(handle, _)) if handle == run3);

        advance(Duration::from_secs(11)).await;
        assert_matches!(event_recv.recv().await, Ok(SectionEvent::RunFinish(handle, _)) if handle == run3);

        runner.quit().await.unwrap();
    }
}
